import json
import logging
from dataclasses import dataclass
from datetime import datetime
from typing import (
    Any,
    Callable,
    Collection,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
)

from pydantic import BaseModel, validator
from snowflake.connector import SnowflakeConnection

from datahub.configuration.datetimes import parse_absolute_time
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.aws.s3_util import make_s3_urn_for_lineage
from datahub.ingestion.source.snowflake.constants import (
    LINEAGE_PERMISSION_ERROR,
    SnowflakeEdition,
)
from datahub.ingestion.source.snowflake.snowflake_config import SnowflakeV2Config
from datahub.ingestion.source.snowflake.snowflake_query import SnowflakeQuery
from datahub.ingestion.source.snowflake.snowflake_report import SnowflakeV2Report
from datahub.ingestion.source.snowflake.snowflake_utils import (
    SnowflakeCommonMixin,
    SnowflakeConnectionMixin,
    SnowflakePermissionError,
    SnowflakeQueryMixin,
)
from datahub.ingestion.source.state.redundant_run_skip_handler import (
    RedundantLineageRunSkipHandler,
)
from datahub.metadata.com.linkedin.pegasus2avro.dataset import (
    FineGrainedLineage,
    UpstreamLineage,
)
from datahub.metadata.schema_classes import DatasetLineageTypeClass, UpstreamClass
from datahub.sql_parsing.sql_parsing_aggregator import (
    ColumnLineageInfo,
    ColumnRef,
    KnownQueryLineageInfo,
    SqlParsingAggregator,
    UrnStr,
)
from datahub.sql_parsing.sqlglot_lineage import DownstreamColumnRef
from datahub.utilities.perf_timer import PerfTimer
from datahub.utilities.time import ts_millis_to_datetime

logger: logging.Logger = logging.getLogger(__name__)

EXTERNAL_LINEAGE = "external_lineage"
TABLE_LINEAGE = "table_lineage"
VIEW_LINEAGE = "view_lineage"


def pydantic_parse_json(field: str) -> classmethod:
    def _parse_from_json(cls: Type, v: Any) -> dict:
        if isinstance(v, str):
            return json.loads(v)
        return v

    return validator(field, pre=True, allow_reuse=True)(_parse_from_json)


class UpstreamColumnNode(BaseModel):
    object_name: str
    object_domain: str
    column_name: str


class ColumnUpstreamJob(BaseModel):
    column_upstreams: List[UpstreamColumnNode]
    query_id: str


class ColumnUpstreamLineage(BaseModel):
    column_name: str
    upstreams: List[ColumnUpstreamJob]


class UpstreamTableNode(BaseModel):
    upstream_object_domain: str
    upstream_object_name: str
    query_id: str


class Query(BaseModel):
    query_id: str
    query_text: str
    start_time: str


class UpstreamLineageEdge(BaseModel):
    DOWNSTREAM_TABLE_NAME: str
    DOWNSTREAM_TABLE_DOMAIN: str
    UPSTREAM_TABLES: Optional[List[UpstreamTableNode]]
    UPSTREAM_COLUMNS: Optional[List[ColumnUpstreamLineage]]
    QUERIES: Optional[List[Query]]

    _json_upstream_tables = pydantic_parse_json("UPSTREAM_TABLES")
    _json_upstream_columns = pydantic_parse_json("UPSTREAM_COLUMNS")
    _json_queries = pydantic_parse_json("QUERIES")


@dataclass(frozen=True)
class SnowflakeColumnId:
    column_name: str
    object_name: str
    object_domain: Optional[str] = None


class SnowflakeLineageExtractor(
    SnowflakeQueryMixin, SnowflakeConnectionMixin, SnowflakeCommonMixin
):
    """
    Extracts Lineage from Snowflake.
    Following lineage edges are considered.

    1. "Table to View" lineage via `snowflake.account_usage.object_dependencies` view + View definition SQL parsing.
    2. "S3 to Table" lineage via `show external tables` query and `snowflake.account_usage.copy_history view.
    3. "View to Table" and "Table to Table" lineage via `snowflake.account_usage.access_history` view (requires Snowflake Enterprise Edition or above)

    Edition Note - Snowflake Standard Edition does not have Access History Feature.
    So it does not support lineage extraction for point 3 edges mentioned above.
    """

    def __init__(
        self,
        config: SnowflakeV2Config,
        report: SnowflakeV2Report,
        dataset_urn_builder: Callable[[str], str],
        redundant_run_skip_handler: Optional[RedundantLineageRunSkipHandler],
        sql_aggregator: SqlParsingAggregator,
    ) -> None:
        self.config = config
        self.report = report
        self.logger = logger
        self.dataset_urn_builder = dataset_urn_builder
        self.connection: Optional[SnowflakeConnection] = None
        self.sql_aggregator = sql_aggregator

        self.redundant_run_skip_handler = redundant_run_skip_handler
        self.start_time, self.end_time = (
            self.report.lineage_start_time,
            self.report.lineage_end_time,
        ) = self.get_time_window()

    def get_time_window(self) -> Tuple[datetime, datetime]:
        if self.redundant_run_skip_handler:
            return self.redundant_run_skip_handler.suggest_run_time_window(
                (
                    self.config.start_time
                    if not self.config.ignore_start_time_lineage
                    else ts_millis_to_datetime(0)
                ),
                self.config.end_time,
            )
        else:
            return (
                (
                    self.config.start_time
                    if not self.config.ignore_start_time_lineage
                    else ts_millis_to_datetime(0)
                ),
                self.config.end_time,
            )

    def get_workunits(
        self,
        discovered_tables: List[str],
        discovered_views: List[str],
    ) -> Iterable[MetadataWorkUnit]:
        if not self._should_ingest_lineage():
            return

        self.connection = self.create_connection()
        if self.connection is None:
            return

        # s3 dataset -> snowflake table
        self._populate_external_upstreams(discovered_tables)

        # snowflake view/table -> snowflake table
        self.populate_table_upstreams(discovered_tables)

        for mcp in self.sql_aggregator.gen_metadata():
            yield mcp.as_workunit()

        if self.redundant_run_skip_handler:
            # Update the checkpoint state for this run.
            self.redundant_run_skip_handler.update_state(
                (
                    self.config.start_time
                    if not self.config.ignore_start_time_lineage
                    else ts_millis_to_datetime(0)
                ),
                self.config.end_time,
            )

    def populate_table_upstreams(self, discovered_tables: List[str]) -> None:
        if self.report.edition == SnowflakeEdition.STANDARD:
            # TODO: use sql_aggregator.add_observed_query to report queries from
            # snowflake.account_usage.query_history and let Datahub generate lineage, usage and operations
            logger.info(
                "Snowflake Account is Standard Edition. Table to Table and View to Table Lineage Feature is not supported."
            )  # See Edition Note above for why
        else:
            with PerfTimer() as timer:
                results = self._fetch_upstream_lineages_for_tables()

                if not results:
                    return

                self.populate_known_query_lineage(discovered_tables, results)
                self.report.table_lineage_query_secs = timer.elapsed_seconds()
            logger.info(
                f"Upstream lineage detected for {self.report.num_tables_with_known_upstreams} tables.",
            )

    def populate_known_query_lineage(
        self,
        discovered_assets: Collection[str],
        results: Iterable[UpstreamLineageEdge],
    ) -> None:
        for db_row in results:
            dataset_name = self.get_dataset_identifier_from_qualified_name(
                db_row.DOWNSTREAM_TABLE_NAME
            )
            if dataset_name not in discovered_assets or not db_row.QUERIES:
                continue

            for query in db_row.QUERIES:
                known_lineage = self.get_known_query_lineage(
                    query, dataset_name, db_row
                )
                if known_lineage and known_lineage.upstreams:
                    self.report.num_tables_with_known_upstreams += 1
                    self.sql_aggregator.add_known_query_lineage(known_lineage, True)
                else:
                    logger.debug(f"No lineage found for {dataset_name}")

    def _create_upstream_lineage_workunit(
        self,
        dataset_name: str,
        upstreams: Sequence[UpstreamClass],
        fine_upstreams: Sequence[FineGrainedLineage],
    ) -> MetadataWorkUnit:
        logger.debug(
            f"Upstream lineage of '{dataset_name}': {[u.dataset for u in upstreams]}"
        )
        if self.config.upstream_lineage_in_report:
            self.report.upstream_lineage[dataset_name] = [u.dataset for u in upstreams]

        upstream_lineage = UpstreamLineage(
            upstreams=sorted(upstreams, key=lambda x: x.dataset),
            fineGrainedLineages=sorted(
                fine_upstreams,
                key=lambda x: (x.downstreams, x.upstreams),
            )
            or None,
        )
        return MetadataChangeProposalWrapper(
            entityUrn=self.dataset_urn_builder(dataset_name), aspect=upstream_lineage
        ).as_workunit()

    def get_known_query_lineage(
        self, query: Query, dataset_name: str, db_row: UpstreamLineageEdge
    ) -> Optional[KnownQueryLineageInfo]:

        if not db_row.UPSTREAM_TABLES:
            return None

        downstream_table_urn = self.dataset_urn_builder(dataset_name)

        known_lineage = KnownQueryLineageInfo(
            query_text=query.query_text,
            downstream=downstream_table_urn,
            upstreams=self.map_query_result_upstreams(
                db_row.UPSTREAM_TABLES, query.query_id
            ),
            column_lineage=(
                self.map_query_result_fine_upstreams(
                    downstream_table_urn,
                    db_row.UPSTREAM_COLUMNS,
                    query.query_id,
                )
                if (self.config.include_column_lineage and db_row.UPSTREAM_COLUMNS)
                else None
            ),
            timestamp=parse_absolute_time(query.start_time),
        )

        return known_lineage

    def _populate_external_upstreams(self, discovered_tables: List[str]) -> None:
        with PerfTimer() as timer:
            self.report.num_external_table_edges_scanned = 0

            self._populate_external_lineage_from_copy_history(discovered_tables)
            logger.info(
                "Done populating external lineage from copy history. "
                f"Found {self.report.num_external_table_edges_scanned} external lineage edges so far."
            )

            self._populate_external_lineage_from_show_query(discovered_tables)
            logger.info(
                "Done populating external lineage from show external tables. "
                f"Found {self.report.num_external_table_edges_scanned} external lineage edges so far."
            )

            self.report.external_lineage_queries_secs = timer.elapsed_seconds()

    # Handles the case for explicitly created external tables.
    # NOTE: Snowflake does not log this information to the access_history table.
    def _populate_external_lineage_from_show_query(
        self, discovered_tables: List[str]
    ) -> None:
        external_tables_query: str = SnowflakeQuery.show_external_tables()
        try:
            for db_row in self.query(external_tables_query):
                key = self.get_dataset_identifier(
                    db_row["name"], db_row["schema_name"], db_row["database_name"]
                )

                if key not in discovered_tables:
                    continue
                if db_row["location"].startswith("s3://"):
                    self.sql_aggregator.add_known_lineage_mapping(
                        downstream_urn=self.dataset_urn_builder(key),
                        upstream_urn=make_s3_urn_for_lineage(
                            db_row["location"], self.config.env
                        ),
                    )
                    self.report.num_external_table_edges_scanned += 1

                self.report.num_external_table_edges_scanned += 1
        except Exception as e:
            logger.debug(e, exc_info=e)
            self.report_warning(
                "external_lineage",
                f"Populating external table lineage from Snowflake failed due to error {e}.",
            )
            self.report_status(EXTERNAL_LINEAGE, False)

    # Handles the case where a table is populated from an external stage/s3 location via copy.
    # Eg: copy into category_english from @external_s3_stage;
    # Eg: copy into category_english from 's3://acryl-snow-demo-olist/olist_raw_data/category_english'credentials=(aws_key_id='...' aws_secret_key='...')  pattern='.*.csv';
    # NOTE: Snowflake does not log this information to the access_history table.
    def _populate_external_lineage_from_copy_history(
        self, discovered_tables: List[str]
    ) -> None:
        query: str = SnowflakeQuery.copy_lineage_history(
            start_time_millis=int(self.start_time.timestamp() * 1000),
            end_time_millis=int(self.end_time.timestamp() * 1000),
            downstreams_deny_pattern=self.config.temporary_tables_pattern,
        )

        try:
            for db_row in self.query(query):
                self._process_external_lineage_result_row(db_row, discovered_tables)
        except Exception as e:
            if isinstance(e, SnowflakePermissionError):
                error_msg = "Failed to get external lineage. Please grant imported privileges on SNOWFLAKE database. "
                self.warn_if_stateful_else_error(LINEAGE_PERMISSION_ERROR, error_msg)
            else:
                logger.debug(e, exc_info=e)
                self.report_warning(
                    "external_lineage",
                    f"Populating table external lineage from Snowflake failed due to error {e}.",
                )
            self.report_status(EXTERNAL_LINEAGE, False)

    def _process_external_lineage_result_row(
        self, db_row: dict, discovered_tables: List[str]
    ) -> None:
        # key is the down-stream table name
        key: str = self.get_dataset_identifier_from_qualified_name(
            db_row["DOWNSTREAM_TABLE_NAME"]
        )
        if key not in discovered_tables:
            return

        if db_row["UPSTREAM_LOCATIONS"] is not None:
            external_locations = json.loads(db_row["UPSTREAM_LOCATIONS"])

            for loc in external_locations:
                if loc.startswith("s3://"):
                    self.sql_aggregator.add_known_lineage_mapping(
                        downstream_urn=self.dataset_urn_builder(key),
                        upstream_urn=make_s3_urn_for_lineage(loc, self.config.env),
                    )
                    self.report.num_external_table_edges_scanned += 1

    def _fetch_upstream_lineages_for_tables(self) -> Iterable[UpstreamLineageEdge]:
        query: str = SnowflakeQuery.table_to_table_lineage_history_v2(
            start_time_millis=int(self.start_time.timestamp() * 1000),
            end_time_millis=int(self.end_time.timestamp() * 1000),
            upstreams_deny_pattern=self.config.temporary_tables_pattern,
            include_view_lineage=self.config.include_view_lineage,
            include_column_lineage=self.config.include_column_lineage,
        )
        try:
            for db_row in self.query(query):
                yield UpstreamLineageEdge.parse_obj(db_row)
        except Exception as e:
            if isinstance(e, SnowflakePermissionError):
                error_msg = "Failed to get table/view to table lineage. Please grant imported privileges on SNOWFLAKE database. "
                self.warn_if_stateful_else_error(LINEAGE_PERMISSION_ERROR, error_msg)
            else:
                logger.debug(e, exc_info=e)
                self.report_warning(
                    "table-upstream-lineage",
                    f"Extracting lineage from Snowflake failed due to error {e}.",
                )
            self.report_status(TABLE_LINEAGE, False)

    def map_query_result_upstreams(
        self, upstream_tables: Optional[List[UpstreamTableNode]], query_id: str
    ) -> List[UrnStr]:
        if not upstream_tables:
            return []
        upstreams: List[UrnStr] = []
        for upstream_table in upstream_tables:
            if upstream_table and upstream_table.query_id == query_id:
                try:
                    upstream_name = self.get_dataset_identifier_from_qualified_name(
                        upstream_table.upstream_object_name
                    )
                    if upstream_name and self._is_dataset_pattern_allowed(
                        upstream_name,
                        upstream_table.upstream_object_domain,
                        is_upstream=True,
                    ):
                        upstreams.append(self.dataset_urn_builder(upstream_name))
                except Exception as e:
                    logger.debug(e, exc_info=e)
        return upstreams

    def map_query_result_fine_upstreams(
        self,
        dataset_urn: str,
        column_wise_upstreams: Optional[List[ColumnUpstreamLineage]],
        query_id: str,
    ) -> List[ColumnLineageInfo]:
        if not column_wise_upstreams:
            return []
        fine_upstreams: List[ColumnLineageInfo] = []
        for column_with_upstreams in column_wise_upstreams:
            if column_with_upstreams:
                try:
                    self._process_add_single_column_upstream(
                        dataset_urn, fine_upstreams, column_with_upstreams, query_id
                    )
                except Exception as e:
                    logger.debug(e, exc_info=e)
        return fine_upstreams

    def _process_add_single_column_upstream(
        self,
        dataset_urn: str,
        fine_upstreams: List[ColumnLineageInfo],
        column_with_upstreams: ColumnUpstreamLineage,
        query_id: str,
    ) -> None:
        column_name = column_with_upstreams.column_name
        upstream_jobs = column_with_upstreams.upstreams
        if column_name and upstream_jobs:
            for upstream_job in upstream_jobs:
                if not upstream_job or upstream_job.query_id != query_id:
                    continue
                fine_upstream = self.build_finegrained_lineage(
                    dataset_urn=dataset_urn,
                    col=column_name,
                    upstream_columns={
                        SnowflakeColumnId(
                            column_name=col.column_name,
                            object_name=col.object_name,
                            object_domain=col.object_domain,
                        )
                        for col in upstream_job.column_upstreams
                    },
                )
                if not fine_upstream:
                    continue
                fine_upstreams.append(fine_upstream)

    def build_finegrained_lineage(
        self,
        dataset_urn: str,
        col: str,
        upstream_columns: Set[SnowflakeColumnId],
    ) -> Optional[ColumnLineageInfo]:
        column_upstreams = self.build_finegrained_lineage_upstreams(upstream_columns)
        if not column_upstreams:
            return None
        column_lineage = ColumnLineageInfo(
            downstream=DownstreamColumnRef(
                table=dataset_urn, column=self.snowflake_identifier(col)
            ),
            upstreams=sorted(column_upstreams),
        )

        return column_lineage

    def build_finegrained_lineage_upstreams(
        self, upstream_columms: Set[SnowflakeColumnId]
    ) -> List[ColumnRef]:
        column_upstreams = []
        for upstream_col in upstream_columms:
            if (
                upstream_col.object_name
                and upstream_col.column_name
                and self._is_dataset_pattern_allowed(
                    upstream_col.object_name,
                    upstream_col.object_domain,
                    is_upstream=True,
                )
            ):
                upstream_dataset_name = self.get_dataset_identifier_from_qualified_name(
                    upstream_col.object_name
                )
                column_upstreams.append(
                    ColumnRef(
                        table=self.dataset_urn_builder(upstream_dataset_name),
                        column=self.snowflake_identifier(upstream_col.column_name),
                    )
                )
        return column_upstreams

    def get_external_upstreams(self, external_lineage: Set[str]) -> List[UpstreamClass]:
        external_upstreams = []
        for external_lineage_entry in sorted(external_lineage):
            # For now, populate only for S3
            if external_lineage_entry.startswith("s3://"):
                external_upstream_table = UpstreamClass(
                    dataset=make_s3_urn_for_lineage(
                        external_lineage_entry, self.config.env
                    ),
                    type=DatasetLineageTypeClass.COPY,
                )
                external_upstreams.append(external_upstream_table)
        return external_upstreams

    def _should_ingest_lineage(self) -> bool:
        if (
            self.redundant_run_skip_handler
            and self.redundant_run_skip_handler.should_skip_this_run(
                cur_start_time=(
                    self.config.start_time
                    if not self.config.ignore_start_time_lineage
                    else ts_millis_to_datetime(0)
                ),
                cur_end_time=self.config.end_time,
            )
        ):
            # Skip this run
            self.report.report_warning(
                "lineage-extraction",
                "Skip this run as there was already a run for current ingestion window.",
            )
            return False
        return True

    def report_status(self, step: str, status: bool) -> None:
        if self.redundant_run_skip_handler:
            self.redundant_run_skip_handler.report_current_run_status(step, status)



import logging
from abc import abstractmethod
from collections import defaultdict
from typing import Any, Dict, List, Union

import sqlalchemy as sa
from sqlalchemy.engine import Connection, Engine
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.engine.row import LegacyRow

from datahub.ingestion.api.closeable import Closeable

logger: logging.Logger = logging.getLogger(__name__)


class DataReader(Closeable):
    @abstractmethod
    def get_sample_data_for_column(
        self, table_id: List[str], column_name: str, sample_size: int = 100
    ) -> list:
        pass

    @abstractmethod
    def get_sample_data_for_table(
        self, table_id: List[str], sample_size: int = 100
    ) -> Dict[str, list]:
        pass


class SqlAlchemyTableDataReader(DataReader):
    @staticmethod
    def create(inspector: Inspector) -> "SqlAlchemyTableDataReader":
        return SqlAlchemyTableDataReader(conn=inspector.bind)

    def __init__(
        self,
        conn: Union[Engine, Connection],
    ) -> None:
        # TODO: How can this use a connection pool instead ?
        self.engine = conn.engine.connect()

    def _table(self, table_id: List[str]) -> sa.Table:
        return sa.Table(
            table_id[-1],
            sa.MetaData(),
            schema=table_id[-2] if len(table_id) > 1 else None,
        )

    def get_sample_data_for_column(
        self, table_id: List[str], column_name: str, sample_size: int = 100
    ) -> list:
        """
        Fetches non-null column values, upto <sample_size> count
        Args:
            table_id: Table name identifier. One of
                        - [<db_name>, <schema_name>, <table_name>] or
                        - [<schema_name>, <table_name>] or
                        - [<table_name>]
            column: Column name
        Returns:
            list of column values
        """

        table = self._table(table_id)
        query: Any
        ignore_null_condition = sa.column(column_name).is_(None)
        # limit doesn't compile properly for oracle so we will append rownum to query string later
        if self.engine.dialect.name.lower() == "oracle":
            raw_query = (
                sa.select([sa.column(column_name)])
                .select_from(table)
                .where(sa.not_(ignore_null_condition))
            )

            query = str(
                raw_query.compile(self.engine, compile_kwargs={"literal_binds": True})
            )
            query += "\nAND ROWNUM <= %d" % sample_size
        else:
            query = (
                sa.select([sa.column(column_name)])
                .select_from(table)
                .where(sa.not_(ignore_null_condition))
                .limit(sample_size)
            )
        query_results = self.engine.execute(query)

        return [x[column_name] for x in query_results.fetchall()]

    def get_sample_data_for_table(
        self, table_id: List[str], sample_size: int = 100
    ) -> Dict[str, list]:
        """
        Fetches table values, upto <sample_size>*1.2 count
        Args:
            table_id: Table name identifier. One of
                        - [<db_name>, <schema_name>, <table_name>] or
                        - [<schema_name>, <table_name>] or
                        - [<table_name>]
        Returns:
            dictionary of (column name -> list of column values)
        """
        column_values: Dict[str, list] = defaultdict(list)
        table = self._table(table_id)

        # Ideally we do not want null values in sample data for a column.
        # However that would require separate query per column and
        # that would be expensiv. To compensate for possibility
        # of some null values in collected sample, we fetch extra (20% more)
        # rows than configured sample_size.
        sample_size = int(sample_size * 1.2)

        query: Any

        # limit doesn't compile properly for oracle so we will append rownum to query string later
        if self.engine.dialect.name.lower() == "oracle":
            raw_query = sa.select([sa.text("*")]).select_from(table)

            query = str(
                raw_query.compile(self.engine, compile_kwargs={"literal_binds": True})
            )
            query += "\nAND ROWNUM <= %d" % sample_size
        else:
            query = sa.select([sa.text("*")]).select_from(table).limit(sample_size)
        query_results = self.engine.execute(query)

        # Not ideal - creates a parallel structure in column_values. Can we use pandas here ?
        for row in query_results.fetchall():
            if isinstance(row, LegacyRow):
                for col, col_value in row.items():
                    column_values[col].append(col_value)

        return column_values

    def close(self) -> None:
        self.engine.close()



import json
import logging
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Tuple, Union

from pydantic import BaseModel, Field, validator
from ruamel.yaml import YAML

from datahub.api.entities.structuredproperties.structuredproperties import (
    AllowedTypes,
    StructuredProperties,
)
from datahub.configuration.common import ConfigModel
from datahub.emitter.mce_builder import (
    make_data_platform_urn,
    make_dataset_urn,
    make_schema_field_urn,
)
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.extractor.schema_util import avro_schema_to_mce_fields
from datahub.ingestion.graph.client import DataHubGraph, get_default_graph
from datahub.metadata.schema_classes import (
    DatasetPropertiesClass,
    MetadataChangeProposalClass,
    OtherSchemaClass,
    SchemaFieldClass,
    SchemaMetadataClass,
    StructuredPropertiesClass,
    StructuredPropertyValueAssignmentClass,
    SubTypesClass,
    UpstreamClass,
)
from datahub.specific.dataset import DatasetPatchBuilder
from datahub.utilities.urns.dataset_urn import DatasetUrn
from datahub.utilities.urns.urn import Urn

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class SchemaFieldSpecification(BaseModel):
    id: Optional[str]
    urn: Optional[str]
    structured_properties: Optional[
        Dict[str, Union[str, float, List[Union[str, float]]]]
    ] = None
    type: Optional[str]
    nativeDataType: Optional[str] = None
    jsonPath: Union[None, str] = None
    nullable: Optional[bool] = None
    description: Union[None, str] = None
    label: Optional[str] = None
    created: Optional[dict] = None
    lastModified: Optional[dict] = None
    recursive: Optional[bool] = None
    globalTags: Optional[dict] = None
    glossaryTerms: Optional[dict] = None
    isPartOfKey: Optional[bool] = None
    isPartitioningKey: Optional[bool] = None
    jsonProps: Optional[dict] = None

    def with_structured_properties(
        self,
        structured_properties: Optional[Dict[str, List[Union[str, float]]]],
    ) -> "SchemaFieldSpecification":
        self.structured_properties = (
            {k: v for k, v in structured_properties.items()}
            if structured_properties
            else None
        )
        return self

    @classmethod
    def from_schema_field(
        cls, schema_field: SchemaFieldClass, parent_urn: str
    ) -> "SchemaFieldSpecification":
        return SchemaFieldSpecification(
            id=Dataset._simplify_field_path(schema_field.fieldPath),
            urn=make_schema_field_urn(
                parent_urn, Dataset._simplify_field_path(schema_field.fieldPath)
            ),
            type=str(schema_field.type),
            nativeDataType=schema_field.nativeDataType,
            nullable=schema_field.nullable,
            description=schema_field.description,
            label=schema_field.label,
            created=schema_field.created.__dict__ if schema_field.created else None,
            lastModified=schema_field.lastModified.__dict__
            if schema_field.lastModified
            else None,
            recursive=schema_field.recursive,
            globalTags=schema_field.globalTags.__dict__
            if schema_field.globalTags
            else None,
            glossaryTerms=schema_field.glossaryTerms.__dict__
            if schema_field.glossaryTerms
            else None,
            isPartitioningKey=schema_field.isPartitioningKey,
            jsonProps=json.loads(schema_field.jsonProps)
            if schema_field.jsonProps
            else None,
        )

    @validator("urn", pre=True, always=True)
    def either_id_or_urn_must_be_filled_out(cls, v, values):
        if not v and not values.get("id"):
            raise ValueError("Either id or urn must be present")
        return v


class SchemaSpecification(BaseModel):
    file: Optional[str]
    fields: Optional[List[SchemaFieldSpecification]]

    @validator("file")
    def file_must_be_avsc(cls, v):
        if v and not v.endswith(".avsc"):
            raise ValueError("file must be a .avsc file")
        return v


class StructuredPropertyValue(ConfigModel):
    value: Union[str, float, List[str], List[float]]
    created: Optional[str]
    lastModified: Optional[str]


class Dataset(BaseModel):
    id: Optional[str]
    platform: Optional[str]
    env: str = "PROD"
    urn: Optional[str]
    description: Optional[str]
    name: Optional[str]
    schema_metadata: Optional[SchemaSpecification] = Field(alias="schema")
    downstreams: Optional[List[str]]
    properties: Optional[Dict[str, str]]
    subtype: Optional[str]
    subtypes: Optional[List[str]]
    structured_properties: Optional[
        Dict[str, Union[str, float, List[Union[str, float]]]]
    ] = None

    @property
    def platform_urn(self) -> str:
        if self.platform:
            return make_data_platform_urn(self.platform)
        else:
            assert self.urn is not None  # validator should have filled this in
            dataset_urn = DatasetUrn.from_string(self.urn)
            return str(dataset_urn.get_data_platform_urn())

    @validator("urn", pre=True, always=True)
    def urn_must_be_present(cls, v, values):
        if not v:
            assert "id" in values, "id must be present if urn is not"
            assert "platform" in values, "platform must be present if urn is not"
            assert "env" in values, "env must be present if urn is not"
            return make_dataset_urn(values["platform"], values["id"], values["env"])
        return v

    @validator("name", pre=True, always=True)
    def name_filled_with_id_if_not_present(cls, v, values):
        if not v:
            assert "id" in values, "id must be present if name is not"
            return values["id"]
        return v

    @validator("platform")
    def platform_must_not_be_urn(cls, v):
        if v.startswith("urn:li:dataPlatform:"):
            return v[len("urn:li:dataPlatform:") :]
        return v

    @classmethod
    def from_yaml(cls, file: str) -> Iterable["Dataset"]:
        with open(file) as fp:
            yaml = YAML(typ="rt")  # default, if not specfied, is 'rt' (round-trip)
            datasets: Union[dict, List[dict]] = yaml.load(fp)
            if isinstance(datasets, dict):
                datasets = [datasets]
            for dataset_raw in datasets:
                dataset = Dataset.parse_obj(dataset_raw)
                yield dataset

    def generate_mcp(
        self,
    ) -> Iterable[Union[MetadataChangeProposalClass, MetadataChangeProposalWrapper]]:
        mcp = MetadataChangeProposalWrapper(
            entityUrn=self.urn,
            aspect=DatasetPropertiesClass(
                description=self.description,
                name=self.name,
                customProperties=self.properties,
            ),
        )
        yield mcp

        if self.schema_metadata:
            if self.schema_metadata.file:
                with open(self.schema_metadata.file, "r") as schema_fp:
                    schema_string = schema_fp.read()
                    schema_metadata = SchemaMetadataClass(
                        schemaName=self.name or self.id or self.urn or "",
                        platform=self.platform_urn,
                        version=0,
                        hash="",
                        platformSchema=OtherSchemaClass(rawSchema=schema_string),
                        fields=avro_schema_to_mce_fields(schema_string),
                    )
                    mcp = MetadataChangeProposalWrapper(
                        entityUrn=self.urn, aspect=schema_metadata
                    )
                    yield mcp

                if self.schema_metadata.fields:
                    for field in self.schema_metadata.fields:
                        field_urn = field.urn or make_schema_field_urn(
                            self.urn, field.id  # type: ignore[arg-type]
                        )
                        assert field_urn.startswith("urn:li:schemaField:")
                        if field.structured_properties:
                            # field_properties_flattened = (
                            #     Dataset.extract_structured_properties(
                            #         field.structured_properties
                            #     )
                            # )
                            mcp = MetadataChangeProposalWrapper(
                                entityUrn=field_urn,
                                aspect=StructuredPropertiesClass(
                                    properties=[
                                        StructuredPropertyValueAssignmentClass(
                                            propertyUrn=f"urn:li:structuredProperty:{prop_key}",
                                            values=prop_value
                                            if isinstance(prop_value, list)
                                            else [prop_value],
                                        )
                                        for prop_key, prop_value in field.structured_properties.items()
                                    ]
                                ),
                            )
                            yield mcp

                    if self.subtype or self.subtypes:
                        mcp = MetadataChangeProposalWrapper(
                            entityUrn=self.urn,
                            aspect=SubTypesClass(
                                typeNames=[
                                    s
                                    for s in [self.subtype] + (self.subtypes or [])
                                    if s
                                ]
                            ),
                        )
                        yield mcp

                    if self.structured_properties:
                        # structured_properties_flattened = (
                        #     Dataset.extract_structured_properties(
                        #         self.structured_properties
                        #     )
                        # )
                        mcp = MetadataChangeProposalWrapper(
                            entityUrn=self.urn,
                            aspect=StructuredPropertiesClass(
                                properties=[
                                    StructuredPropertyValueAssignmentClass(
                                        propertyUrn=f"urn:li:structuredProperty:{prop_key}",
                                        values=prop_value
                                        if isinstance(prop_value, list)
                                        else [prop_value],
                                    )
                                    for prop_key, prop_value in self.structured_properties.items()
                                ]
                            ),
                        )
                        yield mcp

                    if self.downstreams:
                        for downstream in self.downstreams:
                            patch_builder = DatasetPatchBuilder(downstream)
                            assert (
                                self.urn is not None
                            )  # validator should have filled this in
                            patch_builder.add_upstream_lineage(
                                UpstreamClass(
                                    dataset=self.urn,
                                    type="COPY",
                                )
                            )
                            for patch_event in patch_builder.build():
                                yield patch_event

                    logger.info(f"Created dataset {self.urn}")

    @staticmethod
    def extract_structured_properties(
        structured_properties: Dict[str, Union[str, float, List[str], List[float]]]
    ) -> List[Tuple[str, Union[str, float]]]:
        structured_properties_flattened: List[Tuple[str, Union[str, float]]] = []
        for key, value in structured_properties.items():
            validated_structured_property = Dataset.validate_structured_property(
                key, value
            )
            if validated_structured_property:
                structured_properties_flattened.append(validated_structured_property)
        structured_properties_flattened = sorted(
            structured_properties_flattened, key=lambda x: x[0]
        )
        return structured_properties_flattened

    @staticmethod
    def validate_structured_property(
        sp_name: str, sp_value: Union[str, float, List[str], List[float]]
    ) -> Union[Tuple[str, Union[str, float]], None]:
        """
        Validate based on:
        1. Structured property exists/has been created
        2. Structured property value is of the expected type
        """
        urn = Urn.make_structured_property_urn(sp_name)
        with get_default_graph() as graph:
            if graph.exists(urn):
                validated_structured_property = StructuredProperties.from_datahub(
                    graph, urn
                )
                allowed_type = Urn.get_data_type_from_urn(
                    validated_structured_property.type
                )
                try:
                    if not isinstance(sp_value, list):
                        return Dataset.validate_type(sp_name, sp_value, allowed_type)
                    else:
                        for v in sp_value:
                            return Dataset.validate_type(sp_name, v, allowed_type)
                except ValueError:
                    logger.warning(
                        f"Property: {sp_name}, value: {sp_value} should be a {allowed_type}."
                    )
            else:
                logger.error(
                    f"Property {sp_name} does not exist and therefore will not be added to dataset. Please create property before trying again."
                )
        return None

    @staticmethod
    def validate_type(
        sp_name: str, sp_value: Union[str, float], allowed_type: str
    ) -> Tuple[str, Union[str, float]]:
        if allowed_type == AllowedTypes.NUMBER.value:
            return (sp_name, float(sp_value))
        else:
            return (sp_name, sp_value)

    @staticmethod
    def _simplify_field_path(field_path: str) -> str:
        if field_path.startswith("[version=2.0]"):
            # v2 field path
            field_components = []
            current_field = ""
            for c in field_path:
                if c == "[":
                    if current_field:
                        field_components.append(current_field)
                    current_field = ""
                    omit_next = True
                elif c == "]":
                    omit_next = False
                elif c == ".":
                    pass
                elif not omit_next:
                    current_field += c
            if current_field:
                field_components.append(current_field)
            return ".".join(field_components)
        else:
            return field_path

    @staticmethod
    def _schema_from_schema_metadata(
        graph: DataHubGraph, urn: str
    ) -> Optional[SchemaSpecification]:
        schema_metadata: Optional[SchemaMetadataClass] = graph.get_aspect(
            urn, SchemaMetadataClass
        )

        if schema_metadata:
            schema_specification = SchemaSpecification(
                fields=[
                    SchemaFieldSpecification.from_schema_field(
                        field, urn
                    ).with_structured_properties(
                        {
                            sp.propertyUrn: sp.values
                            for sp in structured_props.properties
                        }
                        if structured_props
                        else None
                    )
                    for field, structured_props in [
                        (
                            field,
                            graph.get_aspect(
                                make_schema_field_urn(urn, field.fieldPath),
                                StructuredPropertiesClass,
                            )
                            or graph.get_aspect(
                                make_schema_field_urn(
                                    urn, Dataset._simplify_field_path(field.fieldPath)
                                ),
                                StructuredPropertiesClass,
                            ),
                        )
                        for field in schema_metadata.fields
                    ]
                ]
            )
            return schema_specification
        else:
            return None

    @classmethod
    def from_datahub(cls, graph: DataHubGraph, urn: str) -> "Dataset":
        dataset_properties: Optional[DatasetPropertiesClass] = graph.get_aspect(
            urn, DatasetPropertiesClass
        )
        subtypes: Optional[SubTypesClass] = graph.get_aspect(urn, SubTypesClass)
        structured_properties: Optional[StructuredPropertiesClass] = graph.get_aspect(
            urn, StructuredPropertiesClass
        )
        if structured_properties:
            structured_properties_map: Dict[str, List[Union[str, float]]] = {}
            for sp in structured_properties.properties:
                if sp.propertyUrn in structured_properties_map:
                    assert isinstance(structured_properties_map[sp.propertyUrn], list)
                    structured_properties_map[sp.propertyUrn].extend(sp.values)  # type: ignore[arg-type,union-attr]
                else:
                    structured_properties_map[sp.propertyUrn] = sp.values

        return Dataset(  # type: ignore[call-arg]
            urn=urn,
            description=dataset_properties.description
            if dataset_properties and dataset_properties.description
            else None,
            name=dataset_properties.name
            if dataset_properties and dataset_properties.name
            else None,
            schema=Dataset._schema_from_schema_metadata(graph, urn),
            properties=dataset_properties.customProperties
            if dataset_properties
            else None,
            subtypes=[subtype for subtype in subtypes.typeNames] if subtypes else None,
            structured_properties=structured_properties_map
            if structured_properties
            else None,
        )

    def to_yaml(
        self,
        file: Path,
    ) -> None:
        with open(file, "w") as fp:
            yaml = YAML(typ="rt")  # default, if not specfied, is 'rt' (round-trip)
            yaml.indent(mapping=2, sequence=4, offset=2)
            yaml.default_flow_style = False
            yaml.dump(self.dict(exclude_none=True, exclude_unset=True), fp)
